// Reflow

// Michael Benn, March 2023

// Control software for profile controlled soldering reflow station
// Station supports between one and four hot plates, each of which is under independent PID control
// The station supports one or more temperature profiles, or it can be used in constant temperature mode.
// The device has a rotary encoder and 128 x 64 display for a user interface
// The device supports one thermistor and bi-colour LEDs per hot plate
// The device has one low temperature (40C) cut-off switch for the internal space, but this do
//  not require any software support.

// --------------------------------------------
// Includes
#include <Arduino.h>
#include <Wire.h>               // I2C support
#include <Adafruit_GFX.h>       // Graphics support
#include <Adafruit_SSD1306.h>   // SSD1306 support
#include <U8g2lib.h>
#include "thermProf.h"
//#include "OTA.h"              // We can't use OTA, as that uses WiFi, which disables the OTA :-(


// --------------------------------------------
// Main control definitions
#define PLATES 2          // How many hotplates do we have
#define PROFILES 7        // How many temperature profiles do we have
#define MAXPRSLOTS 10     // What is the maximum number of time slots in a solder profile
#define MAXPROFDURN 600   // What is the longest profile we support, in seconds
#define LEDSPERPLATE 2    // How many LED pairs there are per plate (two for two plates, one for four plates)
#define MAXSSROUT 1500    // Maximum output allowed on the SSR (out of 4096)
#define STATSONSER        // Uncomment this to show stats on the serial port
//#define ENABLEAMBER     // Comment out if your bi-colour LED doesn't do amber very well (like mine!)

// --------------------------------------------
// Structures and data
typedef struct {            // A single entry for a temperature profile
  unsigned int slotSecs;    // The total second count at the end of the profile segment
  unsigned int targetTempC; // What temperature do we want at the end of the time period
} ProfileEntry;

typedef struct {                            // A temperature profile which can be selected
  char *profName;                           // The name of this profile
  byte pidProfile;                          // Which PID profile to use for this profile
  byte slotCount;                           // How many slots there are in this profile
  byte coolingSlot;                         // In which time slot do we start the cooling fan
  int profMinTemp;                          // The minimum temperature for the chart display
  int profMaxTemp;                          // The maximum temperature for the chart display
  int doneTemp;                             // The temperature below which we think the cycle is finished, only applies in last profile slot during cooldown
  ProfileEntry profileSlots[MAXPRSLOTS];    // The array of profile slots
} Profile;

typedef struct {            // The definition of a plate
  char *plateName;          // What we call it
  byte pSSR;                // The pin to drive the solid state relay (SSR)
  byte pwmChannel;          // The PWM channel to use to drive the SSR
  byte pThermistor;         // The pin to read for the thermistor
  byte pRedLED[2];          // The pin(s) the Red LED is on (zero for none)
  byte pGreenLED[2];        // The pin(s) the Green LED is on (zero for none)
  float setPoint;           // The current set point
  float curTemp;            // The last temperature reading in C
  int thermValue;           // The raw thermistor reading off the ADC
  unsigned int ssrOut;      // The current output level to drive the SSR
  unsigned long lastPid;    // The last time we did the last PID action
  float errorPrev;          // The last error value we saw
  float errorInt;           // The last integral value
} Plate;

typedef struct {
  char *pidName;            // The name of this PID profile
  float P;            // The Proportional element 
  float I;            // The Integral element
  float D;            // The Derivative element
  float maxInt;       // The integral limit
  int maxOut;         // The maxiumu output we allow
} PID;


// --------------------------------------------
// Thermistor values
#define THERMRES 6800
#define SERIESRES 100000
#define TEMPSCANS 25    // How many temperature readings we average over
float tempBufs[PLATES][TEMPSCANS];
int thermReadings[PLATES][TEMPSCANS];
int curTempScan = 0;
float tempBufTotals[PLATES];
long thermReadingTotals[PLATES];
bool tempBufsFull = false;
//float R1 = SERIESRES;
//float logR2, R2, T;
float T;
int Tr;
//float c1 = 1.009249522e-03, c2 = 2.378405444e-04, c3 = 2.019202697e-07;
#define LOWESTTHERM 85
#define LOWESTTEMP 23
//float thermProf[4096];

// --------------------------------------------
// Temperature limit definitions
#define WARMC 30      // Below this, LEDs are green, above it they're Amber
#define HOTC 45       // Above this, LEDs go Red
#define VERYHOTC 80   // Above this LEDs flash Red
#define ABSMAXTEMPC 200 // The absolute maximum temperature we allow in the profile, protects against typos!
#define FLASHPER 150    // On and off period for flashing in milliseconds
unsigned long flashTimer = millis();
bool flashOn = false;

// --------------------------------------------
// Define how we affect IO polarity 
#define LEDON HIGH
#define LEDOFF LOW
#define BTNPRESSED LOW

// --------------------------------------------
// PWM variables
#define PWMFREQ 1
#define PWMRES 12
#define PWMOFF 0
#define PWMMAX 4096     // Needs to be changed if PWMRES is changed 
#define PWMACTION 1     // Needs to be set to -1 if the control action needs to be reversed

// --------------------------------------------
// Coms definitions
#define I2CSCL 22
#define I2CSDA 21

// --------------------------------------------
// Cooling fan config
#define pCoolingFan 19
#define COOLOFF 0       // Turns the mid fan off
#define COOLLOW 1024    // The slow speed for the mid fan during a profile or constant run
#define COOLHIGH 4096   // Full speed for the mid fan
#define coolingPwmChannel 3
#define COOLPWMFREQ 100
#define COOLPWMRES 12
#define COOLTHRESH 40

int startCooling; // When we start the cooling fan
bool cooling;     // Are we actively cooling?

// --------------------------------------------
// Screen setup
U8G2_SSD1306_128X64_NONAME_F_HW_I2C u8g2(U8G2_R0, 22, 21);

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels
#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)

static char outstr[100]; // A buffer to use with sprintf

// --------------------------------------------
// Profile mode variables
#define DEFPROFPLATES 1    // How many plates we suggest when profile mode is selected
bool profileLive = false;
byte curProfile = 0;
byte profilePlates = DEFPROFPLATES;
byte curSlot = 0;
int curSecond = 0;
int profileDurn;
bool profileDone = false;
bool profileAbort = false;
float secsPerDispSlot;
int lastPlotX = 0;
int lastSlotSecond;
float avgTemp = 0;
        
unsigned long profileStart;         // The millisecond count when we started profile mode
int setPointDisp[SCREEN_WIDTH];     // Where we store the calculated profile for display 
int tempDisp[SCREEN_WIDTH];         // Where we store the measured values for display 
float setPointsPerSec[MAXPROFDURN]; // Where we store the per second set points


// --------------------------------------------
// Constant mode varables
#define DEFCONSTTEMP 150    // The default temperature we suggest when constant mode is selected
#define DEFCONSTPLATES 1    // How many plates we suggest
#define DEFDURATION 300     // The default duration in seconds
#define MAXDURATION 600     // The default duration in seconds
#define MINCONSTTEMP 0      // Minimum temperature for constant mode
#define MAXCONSTTEMP 220    // Maximum temperature for constant mode
#define RAMPUPDURN 5       // Ramp to desired duration
#define RAMPDOWNDURN 240    // Ramp for cooldown
bool constLive = false;
int constSetPoint = DEFCONSTTEMP;
int constPlates = DEFCONSTPLATES;
int constDurn = DEFDURATION;
bool constDone = false;
bool atTempBeep;            // Whether we've done the at-temperature beep yet
unsigned long constStart;   // The millisecond count when we started constant mode
bool constAbort = false;

// --------------------------------------------
// Main data
Plate plates[PLATES] = {{"Front", 18, 0, 4, {12, 27}, {13, 14}, 0, 0, 0, 0, 0, 0, 0},       // Note that the Back hotplate thermistor is on IO2, which is shown as Therm 3 on the schematic,
                        {"Back", 5, 1, 2, {32, 25}, {33, 26}, 0, 0, 0, 0, 0, 0, 0}          //   that's just to leave a space on the connector between it and the thermo input for the front plate
                     };                                                                     // For four hotplates, utilise IO pins 18, 5, 17 and 16 for the SSRs and 4, 0, 2 and 15 for the thermo inputs.

// Slot definitions are seconds, then temp
Profile profiles[PROFILES] = {{"Chipqk 165 T5", 0, 8, 7, 25, 170, 30, {{5, 45}, {40, 45}, {130, 90}, {220, 130}, {250, 138}, {280, 165}, {310, 138}, {580, 24}}},
                              {"Loc MP218 Fst", 0, 6, 5, 25, 200, 30, {{125, 110}, {190, 130}, {210, 138}, {240, 165}, {270, 138}, {360, 24}}},
                              {"Loc MP218 Slw", 0, 10, 7, 25, 230, 30, {{5, 45}, {40, 45}, {30, 75}, {60, 100}, {120, 125}, {300, 140}, {360, 190}, {420, 215}, {450, 160} , {599, 24}}},
                              {"High Temp T3", 0, 7, 7, 25, 220, 30, {{90, 90}, {180, 130}, {210, 165}, {240, 210}, {270, 170}, {300, 170}, {360, 24}}},
                              {"Demo Quick", 0, 6, 5, 25, 200, 30, {{30, 110}, {45, 130}, {60, 138}, {70, 165}, {80, 138}, {128, 24}}},
                              {"40, 50, 25", 0, 3, 3, 25, 100, 30, {{60, 40}, {120, 50}, {180, 25}}},             // Low temperature profile to aid basic testing
                              {"30, Step 60", 0, 4, 4, 25, 100, 30, {{10, 30}, {11, 60}, {71, 61}, {131, 30}}}    // Step change profiel to aid in tuning PID parameters
                             };

Profile constantProf = {"Constant", 1, 3, 2, 25, 220, 30, {{0, 0}, {0, 0}, {0, 25}}}; // Initial sharp startup to the desired temperature, then desired period, then a ramp down

PID pidProfile[] = {{ "P Only", 24, 0, 0, 20, MAXSSROUT},
                    { "P, I & D", 15, 0, 100, 20, MAXSSROUT}};

// Limits on the SSR output
#define MINOUT 0
#define MAXOUT 1500  // Should be 4096 for maximum control, but be VERY wary about creating massive overshoots on the temperature

int curPlates;
int curPID;

// --------------------------------------------
// Mode management variables
#define MAXMODE 6
#define MAXSUBMODE 6
#define DEFSUBMENUMODE 1
#define MENUMODE 0
#define PROFSTARTMODE 1
#define PROFILEMODE 2
#define CONSTSTARTMODE 3
#define CONSTMODE 4
#define TESTSTARTMODE 5
#define TESTMODE 6
#define CONSTWARN 540       // Show constant shutoff warning after nine minutes
#define CONSTTIMEOUT 600    // Shutoff constant mode after ten minutes unless they confirm they want it to continue

byte curMode = 0; // Which mode we are currently in
char *modeDescs[MAXMODE] = {"Reflow Off", "Profile Selection", "Running Profile", "Constant Setup", "Running Constant"};
byte curSubMode = DEFSUBMENUMODE; // Which submode do we have selected in the current mode
bool subModeActive = false;
bool allowWrap = false;
#define MAINMENUCNT 4
char *mainMenu[MAINMENUCNT] = {"Off", "Profile", "Constant", "Test"};
int modeSubModes[MAXMODE] = {1, 4, 2, 5, 2, 2};

// --------------------------------------------
// Encoder variables
#define pPushButton 35
#define pEncA 36
#define pEncB 34
int encPage = curSubMode;
byte currPage = 0;
#define MAXPAGE 4
byte btmPage;
int topPage;
bool btnPressed = false;
bool menuAction = false;

// --------------------------------------------
// Buzzer confguration
#define pBuzzer 23
#define buzPwmChannel 2

// Now for some buzzer tone definitions

// --------------------------------------------
// Cursor locations, x & y for each pointer in each submode of each mode
byte ptrLocations[MAXMODE][MAXSUBMODE][2] = {
                                              {{0, 32}},                                // Main menu
                                              {{0, 32}, {0, 48}, {0, 64}, {54, 64}},    // Profile start
                                              {{0, 64}, {64, 64}},                      // Profile running
                                              {{0, 32}, {0, 48}, {54, 48}, {0, 64}, {54, 64}},    // Constant start
                                              {{0, 64}, {64, 64}},                      // Constant running
                                              {{0, 64}, {64, 64}}                       // Test mode
                                            };                      
                                            
// --------------------------------------------
// Test mode variables
bool runningTest = false;

// Performance monitoring variables
unsigned long lastPerfCheck = millis();    // Where we hold the time we last checked how many loops we've done per second
int curLoop = 0;
int loopsPerSec = 0;                      

// --------------------------------------------
// Functions

void showLEDs() {
  bool redState;
  bool greenState;

  // Deal with the flash timer globally, not just depending on us being in or out of a flashing state
  if (millis() > (flashTimer + FLASHPER)) {
    flashTimer = millis();
    flashOn = !flashOn; 
  }    

  // Loop around each plate, lighting the LEDs depending on the temperature of that plate
  for (int p = 0; p < PLATES; p++) {
    // Show the LEDs appropriate to the plate temperature. Green for cold, amber for warm, solid red for hot and flashing red for very hot 
    // First work out what state each LED should be in                                                                                                                                             
    float curTemp = plates[p].curTemp;
    greenState = false;
    #ifdef ENABLEAMBER          // Only support amber if that is enabled, otherwise show green up to 'Hot' limit
      if (curTemp < WARMC) {
        // Cold, so show green
        redState = false;
        greenState = true;
      } else if (curTemp < HOTC) {
        // Warm, so show amber (red and green) if warm
        redState = true;
        greenState = true;
      } 
    #else
      if (curTemp < HOTC) {
        // Cold, so show green
        redState = false;
        greenState = true;
      }
    #endif
    else if (curTemp < VERYHOTC) {
      // Hot, so show steady red
      redState = true;
      greenState = false;
    } else {
      // It must be really hot, so flash red
      greenState = false;
      if (flashOn) {
        redState = true;
      } else {
        redState = false;
      }
    }

    // Now that we know what state each LED needs to be, set them accordingly if they are defined
    for (int l = 0; l < LEDSPERPLATE; l++) {
      if (plates[p].pGreenLED[l] != 0) {
        if (greenState) {
          digitalWrite(plates[p].pGreenLED[l], LEDON);
        } else {
          digitalWrite(plates[p].pGreenLED[l], LEDOFF);
        }
      }
      if (plates[p].pRedLED[l] != 0) {
        if (redState) {
          digitalWrite(plates[p].pRedLED[l], LEDON);
        } else {
          digitalWrite(plates[p].pRedLED[l], LEDOFF);
        }
      }
    }
  }
}

float readThermistor(int plate) {
  // Just do an analog read of the thermistors, average and store it, leave all calculations to someone else
  // Do the averaging of the temperature reading
  //int thermValue = 4095 - analogRead(plates[plate].pThermistor);
  int thermValue = analogRead(plates[plate].pThermistor);
  plates[plate].thermValue = thermValue;

  // Use the lookup table instead of a calculatiom for thermo readings
  // If the reading is too low, just use the lowest defined temp instead

  if (thermValue < LOWESTTHERM) {
    T = LOWESTTEMP;
  } else {
    T = thermProf[thermValue - LOWESTTHERM];    // Get the temperature out of the thermistor profile
  }

  if (tempBufsFull) {
    // The temperature buffers are full, so subtract the oldest from the total, add in the new, divide by the sample count and save and return the averaged value
    tempBufTotals[plate] = tempBufTotals[plate] - tempBufs[plate][curTempScan] + T;    // Drop the oldest value and add the current value to the buffer total
    tempBufs[plate][curTempScan]= T;                                                   // Save the new value into the buffer
    T = tempBufTotals[plate] / TEMPSCANS;                                              // Work out the new average, then save it
    plates[plate].curTemp = T;
    // Save the raw readings too
    thermReadingTotals[plate] = thermReadingTotals[plate] - thermReadings[plate][curTempScan] + thermValue;    // Drop the oldest value and add the current value to the buffer total
    thermReadings[plate][curTempScan]= thermValue;                                                   // Save the new value into the buffer
    Tr = thermReadingTotals[plate] / TEMPSCANS;    
    plates[plate].thermValue = Tr;
  } else {
    // We haven't filled the buffer yet, so just record the value and update the buffer total
    tempBufs[plate][curTempScan]= T;
    tempBufTotals[plate] = tempBufTotals[plate] + T;
    thermReadings[plate][curTempScan]= thermValue;
    thermReadingTotals[plate] = thermReadingTotals[plate] + thermValue;
  }

  return plates[plate].curTemp;
}

void ReadEncoder() {
  static uint8_t old_AB = 3;
  static int8_t encVal = 0;
  static const int8_t enc_states[] = {0, -1, 1, 0, 1, 0, 0, -1, -1, 0, 0, 1, 0, 1, -1, 0};

  old_AB <<= 2;
  if (digitalRead(pEncA)) {
    old_AB |= 0x02;
  }
  if (digitalRead(pEncB)) {
    old_AB |= 0x01;
  }

  encVal += enc_states[(old_AB & 0x0f)];

  if (encVal >3 || encVal < -3) {
    // We've moved a full notch, so record and process an action
    //lastAction = millis();
    encPage = encPage + (encVal > 0) - (encVal < 0);
    if (allowWrap) {
      if (encPage >= topPage) {
        encPage = btmPage;
      } else if (encPage < btmPage) {
        encPage = topPage - 1;   
      }
    } else {
      if (encPage >= topPage) {
        encPage = topPage - 1;
      } else if (encPage < btmPage) {
        encPage = btmPage;   
      }
    }
    encVal = 0;
    // Now clear the menuAction flag to show that the encode moved when we were in a yes / no menu item, meaning we want to exit that item, not do it
    menuAction = false;
  }
}

void readThermistors() {
  // Read all of the thermistor values
  for (int p = 0; p < PLATES; p++) {
    plates[p].curTemp = readThermistor(p);

    // If we're doing the last plate, increment the pointer and see if we've wrapped the buffer
    if (p == (PLATES - 1)) {
      if (++curTempScan >= TEMPSCANS) {
        // We have, record that and reset to the start of the buffer
        tempBufsFull = true;
        curTempScan = 0;
      }
    }
  }
}

void showPointers() {
  // Don't show cursor if we are live in constant or profile modes
  if (!profileLive && !constLive) {
    for (int p = 0; p < modeSubModes[curMode]; p++) {
      // Put a > against a chosen submode, an * against an active submode, otherwise blank the character  
      u8g2.drawStr(ptrLocations[curMode][curSubMode][0], ptrLocations[curMode][curSubMode][1], p == curSubMode ? (subModeActive ? "*" : ">") : " ");
    }
  }
}

void setupProfile(Profile * prof, int durn) {
  // First clear the screen and show the background for profile mode including calculating and displaying the setpoint line
  u8g2.clearBuffer();
  u8g2.setDrawColor(1);
  u8g2.setFont(u8g2_font_6x13_tf); 
  u8g2.drawStr(0, 13, prof -> profName); 
  sprintf(outstr, "%3d", durn);
  u8g2.drawStr(110, 13, outstr);
  u8g2.setDrawColor(0);
  u8g2.drawBox(0, 16, 128, 48);
  
  // Put the upper and lower scales in
  u8g2.setDrawColor(1);
  u8g2.setFont(u8g2_font_5x7_tf); 
  sprintf(outstr, "%dC", prof -> profMaxTemp); 
  u8g2.drawStr(0, 23, outstr);
  sprintf(outstr, "%dC", prof -> profMinTemp); 
  u8g2.drawStr(0, 64, outstr);
  u8g2.setFont(u8g2_font_6x13_tf); 
  u8g2.sendBuffer();

  startCooling = 0;

  // Now show the temperature profile
  int slot = 0;
  float targetTempC = prof -> profMinTemp;
  int slotEndSecs = prof -> profileSlots[0].slotSecs;
  int slotSecs = slotEndSecs;
  float rate = (prof -> profileSlots[0].targetTempC - targetTempC) / slotSecs;
  int dispSlot = 0;
  secsPerDispSlot = (float)durn / (float)SCREEN_WIDTH;
  int dispHeight = 40; // The height of the display area we want to draw in
  float degPerPixel = (float)(prof -> profMaxTemp - prof -> profMinTemp) / dispHeight;
  int pointY;
  
  // Scan through the whole profile a second at a time recording the SP, but only draw pixels such as to allow the full profile to be shown within the screen width
  for (int t = 0; t < durn; t++) {
    // Record the set point for this second
    setPointsPerSec[t] = targetTempC;
    
    // See if we've move onto the next slot
    if (t > slotEndSecs) {
      // Yep.  Work out new slot duration and the new rate 
      slot++;
      slotEndSecs = prof -> profileSlots[slot].slotSecs;
      slotSecs =  slotEndSecs - t;
      if (prof -> profileSlots[slot].targetTempC == (targetTempC - rate)) {
        rate = 0; 
      } else {
        rate = (prof -> profileSlots[slot].targetTempC - (targetTempC - rate)) / slotSecs;
      }
      
      // See if this is the last slot.  If so, set the variable to support the program complete buzzer
      if (slot == (prof -> slotCount - 1)) {
        lastSlotSecond = t;  
      }

      // Check to see if we need to start the cooling fan
      if ((startCooling == 0) && (slot == prof -> coolingSlot)) {
        startCooling = t;
      }
    }
    
    // See if we need to record and draw a new SP dot 
    if (((float)t / secsPerDispSlot) > dispSlot) {
      // Yep, work out where
      pointY = (int)((float)SCREEN_HEIGHT - 8 - ((targetTempC - prof -> profMinTemp) / degPerPixel));
      setPointDisp[dispSlot] = pointY;
      
      // Draw the point
      u8g2.drawPixel(dispSlot, pointY);
      
      dispSlot++;
    }

    // Work out the new setpoint
    targetTempC = targetTempC + rate;
  }
}

int doPid(int p, bool show) {

  // Do the PID control for an individual plate

  // Find out how long since we last did a control action, in seconds and then resave the new time
  unsigned long dt = (millis() - plates[p].lastPid);
  plates[p].lastPid = millis();

  // Get the current error
  float errorVal = plates[p].setPoint - plates[p].curTemp;
  
  // Calculate the derivative 
  float errorDer = (errorVal - plates[p].errorPrev) / (dt / 1000.0);

  // Calculate the integral term
  plates[p].errorInt = plates[p].errorInt + (errorVal * (dt / 1000.0));
   
  // Limit the intergral here!
  if (plates[p].errorInt > pidProfile[curPID].maxInt) {
    plates[p].errorInt = pidProfile[curPID].maxInt;
  }

  if (plates[p].errorInt < (-1 * pidProfile[curPID].maxInt)) {
    plates[p].errorInt = -pidProfile[curPID].maxInt;
  }

  // Calculate and limit the output
  float opVal = (pidProfile[curPID].P * errorVal) + (pidProfile[curPID].I * plates[p].errorInt) + (pidProfile[curPID].D * errorDer);

  opVal = (opVal < pidProfile[curPID].maxOut ? opVal : pidProfile[curPID].maxOut);
  
  opVal = (opVal > MINOUT ? opVal : MINOUT);

  int outVal = (int) opVal;      // * PWMACTION;

  // Remember the last error
  plates[p].errorPrev = errorVal;

  // Save what the output should be set to
  if ((errorVal != errorVal) || (plates[p].errorInt != plates[p].errorInt) || (errorDer != errorDer)) {
    // If the errorvalue, integral or diff value are a NaN, turn off the heaters
    plates[p].ssrOut = PWMOFF;
    ledcWrite(plates[p].pwmChannel, PWMOFF); 
    Serial.println("Invalid calculation, heaters off!");
  } else {
    // If we are cooling, don't turn the outputs on
    if (cooling) { 
      plates[p].ssrOut = 0;
    } else {
      plates[p].ssrOut = outVal;
    }
  }
  
  if (show) {
/*
      Serial.print("V: ");
      Serial.print(plates[p].curTemp);
      Serial.print(", SP: ");
      Serial.print(plates[p].setPoint);
      Serial.print(", E: ");
      Serial.print(errorVal);
      Serial.print(", I: ");
      Serial.print(plates[p].errorInt);
      Serial.print(", D: ");
      Serial.print(errorDer);
      Serial.print(", OV: ");
      Serial.println(opVal  * 100.0 / MAXOUT);
      Serial.print(", OP: ");
      Serial.println(outVal);
*/
    
}
  
  return outVal;
}

// --------------------------------------------
void setup() {
  Serial.begin(115200);

  Serial.println("Starting Reflow");

  // Set up OTA uploads
  //setupOTA("Reflow", mySSID, myPASSWORD);
  //Serial.println("OTA Configured");

  // Initialise the IO
  for (int i = 0; i < PLATES; i++) {
    Serial.print("Initialising ");
    Serial.print(plates[i].plateName);
    Serial.println(" plate");
    
    // First the LEDs, if they are defined, flashing each one to show it is working
    for (int l = 0; l < LEDSPERPLATE; l++) {
      if (plates[i].pRedLED[l] != 0) {
        pinMode(plates[i].pRedLED[l], OUTPUT);
        digitalWrite(plates[i].pRedLED[l], LEDON);
        delay(500);
        digitalWrite(plates[i].pRedLED[l], LEDOFF);
      }
      if (plates[i].pGreenLED[l] != 0) {
        pinMode(plates[i].pGreenLED[l], OUTPUT);
        digitalWrite(plates[i].pGreenLED[l], LEDON);
        delay(500);
        digitalWrite(plates[i].pGreenLED[l], LEDOFF);
      }
    }  
    
    // Then the thermistor input
    pinMode(plates[i].pThermistor, INPUT);
    //setupThermProfile();

   // And finally the SSRs
    ledcAttachPin(plates[i].pSSR, plates[i].pwmChannel);
    ledcSetup(plates[i].pwmChannel, PWMFREQ, PWMRES);  /// Connect it to the PWM channel
    ledcWrite(plates[i].pwmChannel, PWMOFF); 
  }
  Serial.println("Plates initialsed");

  // Setup the cooling fan output
  ledcAttachPin(pCoolingFan, coolingPwmChannel);
  ledcSetup(coolingPwmChannel, COOLPWMFREQ, COOLPWMRES);  /// Connect it to the PWM channel
  ledcWrite(coolingPwmChannel, COOLOFF); 
  cooling = false;
    
  // Set up the buzzer
  ledcAttachPin(pBuzzer, buzPwmChannel);

  // Setup the encoder. As we are using inputs between 34 and 39 and these have no internal pullups, we need to use external ones
  // So, if we use different IO pins, we could use pinMode(pin, INPUT_PULLUP)
  pinMode(pPushButton, INPUT);
  pinMode(pEncA, INPUT);
  pinMode(pEncB, INPUT);

  // Add the encoder interrupt handlers
  attachInterrupt(digitalPinToInterrupt(pEncA), ReadEncoder, CHANGE);                                                   
  attachInterrupt(digitalPinToInterrupt(pEncB), ReadEncoder, CHANGE);
  Serial.println("Encoder initialsed");

  // Setup the main menu for first run
  curMode = 0;
  curSubMode = 0; //DEFSUBMENUMODE;
  encPage = DEFSUBMENUMODE;  
  btmPage = 0;
  topPage = MAINMENUCNT;                                                    
  allowWrap = true;

  // Initialise the screen and set the font
  u8g2.begin();
  u8g2.setFont(u8g2_font_6x13_tf); 
}

// --------------------------------------------
void loop() {
// Whatever we do, we should always read the current temperature of the plates and show on the LEDs
  readThermistors();
  showLEDs();

  bool show = false;

  // See if it's time to check the performance loop count
  if (millis() >= lastPerfCheck + 1000) {
    loopsPerSec = curLoop;
    lastPerfCheck = millis();
    curLoop = 0;
    show = true;
    #ifdef STATSONSER
      sprintf(outstr,"SP1:%d T1:%d OP1%%:%d SP2:%d T2:%d OP2%%:%d", (int)plates[0].setPoint, (int)plates[0].curTemp,  plates[0].ssrOut * 100 / MAXOUT, (int)plates[1].setPoint, (int)plates[1].curTemp,  plates[1].ssrOut * 100 / MAXOUT);
    #endif
    Serial.println(outstr);
  } else {
    curLoop++;
  }
  
  // If either the profile or constant modes are live, go and do the control actions first
  if (profileLive || constLive) {
    for (int p = 0 ; p < curPlates; p++) {
      doPid(p, show);
      ledcWrite(plates[p].pwmChannel, plates[p].ssrOut);  
    }
  } else if (curMode == TESTMODE  || curMode == TESTSTARTMODE) {
    // Do anything that needs doing in test mode
  } else {
    // Just in normal menu / setup mode, so show default screen
      // Now for the main meat of the looping process
      // If we aren't at the end of a profile or constant run, show the menu mode
      if (!profileDone && !profileAbort) {
        u8g2.clearBuffer();
        u8g2.drawStr(0, 14, modeDescs[curMode]); 
        // Put in the cursor, which shows what the user is selecting or changing
        showPointers(); 
      } 
  
      // If any of the plates are hot, turn on the cooling fan
      bool coolReqd = false;

      for (int p = 0; p < PLATES; p++) {
        if (plates[p].curTemp >= COOLTHRESH) {
          coolReqd = true;
        }
      }

      // Turn the cooling fan on / off as required
      if (coolReqd) {
        ledcWrite(coolingPwmChannel, COOLHIGH);
      } else {
        ledcWrite(coolingPwmChannel, COOLOFF);        
      }
  }

  
  // Now read the encoder button
  if (digitalRead(pPushButton) == BTNPRESSED) {
    btnPressed = true;
    Serial.println("Btn Pressed");
    delay(50);  // Debounce the press
  } else {
    btnPressed = false;
  }


     
  // Now do whatever we need to for the current mode
  switch (curMode) {
    case MENUMODE:
      // Just show the options and allow selection with button
      btmPage = 0; 
      topPage = MAINMENUCNT;

      // Show the currently selected mode
      u8g2.drawStr(10, 32, mainMenu[encPage]); 

      // Show the loops per sec
      sprintf(outstr,"%d", loopsPerSec);
      u8g2.drawStr(100, 32, outstr); 

      // Show the current plate temps
      for (int p = PLATES - 1; p >= 0; p--) {
        sprintf(outstr, "%6s: %dC  %d", plates[p].plateName, (int)plates[p].curTemp, (int)plates[p].thermValue);
        u8g2.drawStr(20, 36 + (12 * (PLATES - p)), outstr);   
      }
      u8g2.sendBuffer(); 

      // Action the button if pressed
      if (btnPressed) {
        // Make sure the new mode starts at the first submode
        curSubMode = 0;
        // Then set the new mode and the encoder limits and start value
        switch (encPage) {
          case 0: // Off mode

            break;
          case 1: // Profile mode
            curMode = PROFSTARTMODE;
            subModeActive = false;
            curProfile = 0;
            profilePlates = DEFPROFPLATES;
            btmPage = 0;
            topPage = modeSubModes[curMode];
            encPage = 2;
            break;
          case 2: // Constant temp mode
            curMode = CONSTSTARTMODE;
            subModeActive = false;
            constSetPoint = DEFCONSTTEMP;
            constPlates = DEFCONSTPLATES;
            btmPage = 0;
            topPage = modeSubModes[curMode];
            encPage = 0;
            break;
          case 3: // Test mode
            curMode = TESTSTARTMODE;
            subModeActive = false;
            btmPage = 0;
            topPage = modeSubModes[curMode];
            encPage = 0;
            break;
          }
          // Wait for the button to be released before proceeeding
          while (digitalRead(pPushButton) == BTNPRESSED) {
            delay(100);
          }
      }
      break;
    case PROFSTARTMODE: // Profile selection mode, ask for profile name and number of plates
      // See if they've pressed the button 
      if (btnPressed) {
        if (subModeActive) {  
          // They pressed the button while the sub mode is active, so see if we  need to do anythng
          switch (curSubMode) {
            case 0:   // Profile selection  
            case 1:   // Plate count
              subModeActive = false;  // The user has change the value as required, so flip back to submode selection and move to the next submode
              curSubMode++;
              encPage = curSubMode;
              if (curSubMode >= modeSubModes[curMode]) {
                curSubMode = 0;
                encPage = 0;
              }
              btmPage = 0;
              topPage = modeSubModes[curMode];
              break;
            case 2:   // Go!
              curMode = PROFILEMODE;
              curSubMode = 0;
              btmPage = 0;
              topPage = modeSubModes[curMode];
              encPage = 0;
              subModeActive = false;
              // Start the temperature ramp now
              //++profilePlates;    // Inc the plate count as it is currently 0 for 1 plate
              curSlot = 0;
              curSecond = 0;
              lastPlotX = 0;
              profileLive = true;
              profileStart = millis();
              profileDurn = profiles[curProfile].profileSlots[profiles[curProfile].slotCount - 1].slotSecs;
              setupProfile(&profiles[curProfile], profileDurn);
              // Start the cooling fan on low, just to keep the ABS coolish
               // And the cooling fan output
              ledcWrite(coolingPwmChannel, COOLLOW); 
              curPlates = profilePlates;
              curPID = profiles[curProfile].pidProfile;
              for (int p = 0; p < PLATES; p++) {
                plates[p].lastPid = millis();
                plates[p].errorPrev = 0;
                plates[p].errorInt = 0;
              }
              Serial.println("Profile started");
              Serial.print("Plates: ");
              Serial.println(profilePlates);
              break;
            case 3: // Exit
              curMode = 0;
              btmPage = 0;
              topPage = MAXMODE;
              curSubMode = 0;
              encPage = 0;
              subModeActive = false;
              break;
          }
        } else {
          // The submode isn't already active, so make it so 
          subModeActive = true; 
          switch (curSubMode) {
            case 0:   // Profile selection  
              encPage = curProfile;
              btmPage = 0;
              topPage = PROFILES;
              break;
            case 1:   // Plate count
              encPage = profilePlates;
              btmPage = 1;
              topPage = PLATES + 1;
              break;
            case 2:   // Go
            case 3:   // Exit
              menuAction = true;
              break;
          }
        }
        // Wait for the button to be released before proceeeding
        while (digitalRead(pPushButton) == BTNPRESSED) {
          delay(100);
        }
      } else {
        // No button press, so update the appropriate variable with the current encode value
        if (subModeActive) {
          // The submode is active so update the underlying value
          switch (curSubMode) {
            case 0: // Profile
              curProfile = encPage;
              break;
            case 1: // Plate count
              profilePlates = encPage;
              break;
            case 2: // Go
            case 3: // Exit
              if (!menuAction) {
                subModeActive = false;  // They selected go or exit, then turned the knob, so just go back into sub selection mode at the top
                menuAction = false;
                curSubMode = 0;
                encPage = 0;
              }
              break;
          }
        } else {
          // The submode is not active, so change the submode number based on the encoder
          curSubMode = encPage;
        }
      }

      if (!profileLive) {
        // Show the current selections and the go / exit menu
        sprintf(outstr,"Prof: %s",profiles[curProfile].profName);
        u8g2.drawStr(10, 32, outstr); 
        sprintf(outstr,"Plates: %d", profilePlates);
        u8g2.drawStr(10, 48, outstr); 
        u8g2.drawStr(10, 64, "Go!"); 
        u8g2.drawStr(64, 64, "Exit"); 
        u8g2.sendBuffer(); 
      }
      break;
    case PROFILEMODE: 
      // See if they've pressed the button 
      if (btnPressed) {
        if (profileDone || profileAbort) {  
          // They pressed the button when the profile is finished or has been aborted, so just return to main menu
          //ledcWrite(coolingPwmChannel, COOLOFF); 
          profileDone = false;
          profileAbort = false;
          curMode = 0;
          btmPage = 0;
          topPage = MAXMODE;
          curSubMode = 0;
          encPage = 0;
          subModeActive = false;
          u8g2.setDrawColor(1);
          u8g2.clearBuffer();
        } else {
          // Initially, if they press the button mid-profile, that means an emergency stop, so just turn off the SSRs for all plates
          profileAbort = true;
          u8g2.setDrawColor(0);
          u8g2.drawBox(35, 26, 52, 15);
          u8g2.setDrawColor(1);
          u8g2.drawStr(37, 40, "ABORTED!");
          u8g2.setDrawColor(1);

          // As they've aborted, turn the cooling fan on to get the termperature down ASAP
          ledcWrite(coolingPwmChannel, COOLHIGH); 
          cooling = true;
          
          // Then turn the plates off
          for (int p = 0; p < PLATES; p++) {
            plates[p].setPoint = 0;                   // Change the set point to zero so that any PID doesn't fight us
            plates[p].ssrOut = PWMOFF;               
            ledcWrite(plates[p].pwmChannel, PWMOFF);
          }
        }
        // Wait for the button to be released before proceeeding
        while (digitalRead(pPushButton) == BTNPRESSED) {
          delay(100);
        }
      } else {
        // No button press, so update the appropriate variable with the current encode value
        curSubMode = encPage;
      }

      if (!profileDone && !profileAbort) {
        // We're still running, so keep updating the screen and setpoints
        curSecond = (millis() - profileStart) / 1000;
        // Keep running while we're within the profile, as long as we're not in the last phase and the temperature has fallen below the completed temp
        if ((curSecond < profileDurn) && (!((curSecond >= lastSlotSecond) && (avgTemp < profiles[curProfile].doneTemp)))) {
          // See if it's time to start the cooling fan
          if (curSecond == startCooling) {
            // Yep, so start the fan, then reset the start point so that we don't do it again.
            ledcWrite(coolingPwmChannel, COOLHIGH); 
            startCooling = 0;
            cooling = true;
          }

          // Show the remaining duration
          u8g2.setDrawColor(1);
          u8g2.setFont(u8g2_font_6x13_tf); 
          u8g2.drawStr(0, 13, profiles[curProfile].profName);
          int remSecs = profileDurn - curSecond - 1;
          int chars = sprintf(outstr, "%3d", remSecs);
          u8g2.drawStr(110, 13, outstr);
          if (remSecs < 100) {
            u8g2.setDrawColor(0);
            u8g2.drawBox(110, 0, remSecs < 10 ? 12 : 6, 13);
            u8g2.setDrawColor(1);
          }

          // Update the setpoints per plate and total the temp across all live plates
          avgTemp = 0;
          for (int p = 0; p < curPlates; p++) {
            // Set the setpoint for each 
            plates[p].setPoint = setPointsPerSec[curSecond];
            // Tot up the total temperature for the live plates
            avgTemp = avgTemp + plates[p].curTemp;
          }

          // Work out the average temp, then plot it on the chart if it's time to do that
          avgTemp = avgTemp / profilePlates;
          int xPoint = (float)curSecond / (float)secsPerDispSlot;
          if (lastPlotX != xPoint) {
            // Draw the line
            u8g2.setDrawColor(2);
            float lineSize = ((avgTemp - profiles[curProfile].profMinTemp) / (profiles[curProfile].profMaxTemp - profiles[curProfile].profMinTemp)) * 40;
            u8g2.drawLine(xPoint, 56, xPoint, 56 - (int)lineSize);

            lastPlotX = xPoint;
          }
            
          // Then the show current values
          u8g2.setDrawColor(0);
          u8g2.setFont(u8g2_font_5x7_tf);
          u8g2.drawBox(20, 56, 108, 8); 
          u8g2.setDrawColor(1);
          sprintf(outstr,"SP:%dC F:%dC R:%dC  ", (int)plates[0].setPoint, (int)plates[0].curTemp, (int)plates[1].curTemp);
          u8g2.drawStr(22, 64, outstr);          
          u8g2.setDrawColor(1);
        } else {
          profileDone = true;
          u8g2.setDrawColor(0);
          u8g2.drawBox(35, 27, 52, 15);
          u8g2.setDrawColor(1);
          u8g2.drawStr(37, 40, "COMPLETE");        
        }
      } else {
        // See if this is the first time here
        if (profileLive) {
          // Yep, flag profiel as no longer live
          profileLive = false;
          
          // Right, the profile is done, so set the set point to zero, switch off the SRRs and show we're done
          ledcWriteTone(buzPwmChannel , 600);
          for (int p = 0; p < PLATES; p++) {
            plates[p].setPoint = 0;
            ledcWrite(plates[p].pwmChannel, PWMOFF);
          }
          delay(2000);
          ledcWriteTone(buzPwmChannel , 0);

          // We'll now wait 'til they press the button
        } 
      }
      u8g2.sendBuffer();       
      break;
    case CONSTSTARTMODE:  // Constant temperature mode, ask for target temperature, number of plates and duration
      // See if they've pressed the button 
      if (btnPressed) {
        if (subModeActive) {  
          // They pressed the button while the sub mode is active, so see if we  need to do anything
          switch (curSubMode) {
            case 0:   // Target temp selection  
            case 1:   // Plate count
            case 2:   // Duration
              subModeActive = false;  // The user has change the value as required, so flip back to submode selection and move to the next submode
              curSubMode++;
              allowWrap = true;
              encPage = curSubMode;
              if (curSubMode >= modeSubModes[curMode]) {
                curSubMode = 0;
                encPage = 0;
              }
              btmPage = 0;
              topPage = modeSubModes[curMode];
              break;
            case 3:   // Go!
              curMode = CONSTMODE;
              curSubMode = 0;
              btmPage = 0;
              topPage = modeSubModes[curMode];
              encPage = 0;
              subModeActive = false;
              
              // Start the temperature ramp now
              curSlot = 0;
              curSecond = 0;
              lastPlotX = 0;
              constLive = true;
              constStart = millis();
              
              // Now that we know what the user wants, put the required temp and duration into the constant profile slots
              constantProf.profileSlots[0].slotSecs = constantProf.profileSlots[1].slotSecs + RAMPUPDURN;
              constantProf.profileSlots[0].targetTempC = constSetPoint;
              constantProf.profileSlots[1].slotSecs = constantProf.profileSlots[1].slotSecs + RAMPUPDURN + constDurn;
              constantProf.profileSlots[1].targetTempC = constSetPoint;
              
              // Now update the final slot time based on the required duration
              constantProf.profileSlots[2].slotSecs = constantProf.profileSlots[1].slotSecs + RAMPDOWNDURN; 
              constDurn = constantProf.profileSlots[constantProf.slotCount - 1].slotSecs;
              setupProfile(&constantProf, constDurn);
              //delay(5000);
              
              // Start the cooling fan on low, just to keep the PLA coolish
              ledcWrite(coolingPwmChannel, COOLLOW); 
    
              // Set the initial plate conditions   
              for (int p = 0; p < PLATES; p++) {
                plates[p].lastPid = millis();
                plates[p].errorPrev = 0;
                plates[p].errorInt = 0;
              }

              curPlates = constPlates + 1;
              curPID = constantProf.pidProfile;
              
              Serial.println("Constant Run started");
              Serial.print("Plates: ");
              Serial.println(curPlates);
              break;
            case 4: // Exit
              curMode = 0;
              btmPage = 0;
              topPage = MAXMODE;
              curSubMode = 0;
              encPage = 0;
              subModeActive = false;
              break;
          }
        } else {
          // The submode isn't already active, so make it so 
          subModeActive = true; 
          allowWrap = true;
          switch (curSubMode) {
            case 0:   // Set point  
              encPage = constSetPoint;
              btmPage = MINCONSTTEMP;
              topPage = MAXCONSTTEMP;
              allowWrap = false;
              break;
            case 1:   // Plate count
              encPage = constPlates;
              btmPage = 0;
              topPage = PLATES;
              break;
            case 2:   // Duration
              encPage = constDurn;
              btmPage = 0;
              topPage = MAXDURATION;
              break;
            case 3:   // Go
            case 4:   // Exit
              menuAction = true;
              break;
          }
        }
        // Wait for the button to be released before proceeeding
        while (digitalRead(pPushButton) == BTNPRESSED) {
          delay(100);
        }
      } else {
        // No button press, so update the appropriate variable with the current encode value
        if (subModeActive) {
          // The submode is active so update the underlying value
          switch (curSubMode) {
            case 0: // Target temp
              constSetPoint = encPage;

              Serial.print("SP: ");
              Serial.println(constSetPoint);
              break;
            case 1: // Plate count
              constPlates = encPage;
              break;
            case 2: // Duration
              constDurn = encPage;
              break;
            case 3: // Go
            case 4: // Exit
              if (!menuAction) {
                subModeActive = false;  // They selected go or exit, then turned the knob, so just go back into sub selection mode at the top
                menuAction = false;
                curSubMode = 0;
                encPage = 0;
              }
              break;
          }
        } else {
          // The submode is not active, so change the submode number based on the encoder
          curSubMode = encPage;
        }
      }

      if (!constLive) {
        // Show the current selections and the go / exit menu
        sprintf(outstr,"Target: %d", constSetPoint);
        u8g2.drawStr(10, 32, outstr); 
        sprintf(outstr,"Plts: %d", constPlates + 1);
        u8g2.drawStr(10, 48, outstr); 
        sprintf(outstr,"Durn: %d", constDurn + 1);
        u8g2.drawStr(64, 48, outstr); 
        u8g2.drawStr(10, 64, "Go!"); 
        u8g2.drawStr(64, 64, "Exit"); 
        u8g2.sendBuffer(); 
      }
      break;
    case CONSTMODE:
      // See if they've pressed the button 
      if (btnPressed) {
        if (constDone || constAbort) {  
          // They pressed the button when the constant run is finished or has been aborted, so just return to main menu
          //ledcWrite(coolingPwmChannel, COOLOFF); 
          constDone = false;
          constAbort = false;
          curMode = 0;
          btmPage = 0;
          topPage = MAXMODE;
          curSubMode = 0;
          encPage = 0;
          subModeActive = false;
          u8g2.setDrawColor(1);
          u8g2.clearBuffer();
        } else {
          // Initially, if they press the button mid-constant run, that means an emergency stop, so just turn off the SSRs for all plates
          constAbort = true;
          u8g2.setDrawColor(0);
          u8g2.drawBox(35, 26, 52, 15);
          u8g2.setDrawColor(1);
          u8g2.drawStr(37, 40, "ABORTED!");
          u8g2.setDrawColor(1);

          // As they've aborted, turn the cooling fan on to get the termperature down ASAP
          ledcWrite(coolingPwmChannel, COOLHIGH); 
          cooling = true;
          
          // Then turn the plates off
          for (int p = 0; p < PLATES; p++) {
            plates[p].setPoint = 0;                   // Change the set point to zero so that any PID doesn't fight us
            plates[p].ssrOut = 0;                   // Change the set point to zero so that any PID doesn't fight us
            ledcWrite(plates[p].pwmChannel, PWMOFF);
          }
        }
        // Wait for the button to be released before proceeeding
        while (digitalRead(pPushButton) == BTNPRESSED) {
          delay(100);
        }
      } else {
        // No button press, so update the appropriate variable with the current encode value
        curSubMode = encPage;
      }

      if (!constDone && !constAbort) {
        // We're still running, so keep updating the screen and setpoints
        curSecond = (millis() - constStart) / 1000;
        
        // Keep running while we're within the constant run, as long as we're not in the last phase and the temperature has fallen below the completed temp
        if ((curSecond < constDurn) && (!((curSecond >= lastSlotSecond) && (avgTemp < constantProf.doneTemp)))) {
          // See if it's time to start the cooling fan
          if (curSecond == startCooling) {
            // Yep, so start the fan, then reset the start point so that we don't do it again.
            ledcWrite(coolingPwmChannel, COOLHIGH); 
            startCooling = 0;
            cooling = true;
          }

          // Show the remaining duration
          u8g2.setDrawColor(1);
          u8g2.setFont(u8g2_font_6x13_tf); 
          u8g2.drawStr(0, 13, "Constant Run");
          int remSecs = constDurn - curSecond - 1;
          int chars = sprintf(outstr, "%3d", remSecs);
          u8g2.drawStr(110, 13, outstr);
          if (remSecs < 100) {
            u8g2.setDrawColor(0);
            u8g2.drawBox(110, 0, remSecs < 10 ? 12 : 6, 13);
            u8g2.setDrawColor(1);
          }

          // Update the setpoints per plate and total the temp across all live plates
          avgTemp = 0;
          for (int p = 0; p < curPlates; p++) {
            // Set the setpoint for each 
            plates[p].setPoint = setPointsPerSec[curSecond];
            // Tot up the total temperature for the live plates
            avgTemp = avgTemp + plates[p].curTemp;
          }

          // Work out the average temp, then plot it on the chart if it's time to do that
          avgTemp = avgTemp / curPlates;
          int xPoint = (float)curSecond / (float)secsPerDispSlot;
          if (lastPlotX != xPoint) {
            // Draw the line
            u8g2.setDrawColor(2);
            float lineSize = ((avgTemp - 25) / (220 - 25)) * 40;
            u8g2.drawLine(xPoint, 56, xPoint, 56 - (int)lineSize);
            
            lastPlotX = xPoint;
          }

          // Then the show current values
          u8g2.setDrawColor(0);
          u8g2.setFont(u8g2_font_5x7_tf);
          
          u8g2.drawBox(20, 56, 108, 8); 
          u8g2.setDrawColor(1);
          sprintf(outstr,"SP:%dC F:%dC R:%dC  ", (int)plates[0].setPoint, (int)plates[0].curTemp, (int)plates[1].curTemp);
          u8g2.drawStr(22, 64, outstr);

          u8g2.setDrawColor(1);
        } else {
          constDone = true;
          u8g2.setDrawColor(0);
          u8g2.drawBox(35, 27, 52, 15);
          u8g2.setDrawColor(1);
          u8g2.drawStr(37, 40, "COMPLETE");        
        }
      } else {
        // See if this is the first time here
        if (constLive) {
          // Yep, flag constant run as no longer live
          constLive = false;
          
          // Right, the constant run is done, so set the set point to zero, switch off the SRRs and show we're done
          ledcWriteTone(buzPwmChannel , 600);
          for (int p = 0; p < PLATES; p++) {
            plates[p].setPoint = 0;
            ledcWrite(plates[p].pwmChannel, PWMOFF);
          }
          delay(2000);
          ledcWriteTone(buzPwmChannel , 0);

          // We'll now wait 'til they press the button
        } 
      }
      u8g2.sendBuffer();       
      break;
    case TESTSTARTMODE:
      encPage = 0;
      btmPage = 0;
      topPage = MAXSSROUT;
      curMode = TESTMODE;
      allowWrap = false;
      u8g2.clearDisplay();
      u8g2.setFont(u8g2_font_6x10_tf);
      sprintf(outstr,"Test is");
      u8g2.drawStr(0, 16, outstr);
      sprintf(outstr,"PWM:");
      u8g2.drawStr(0, 32, outstr);
      sprintf(outstr,"Rear :");
      u8g2.drawStr(0, 48, outstr);
      sprintf(outstr,"Front:");
      u8g2.drawStr(0, 64, outstr);
      u8g2.sendBuffer();    

      break;
    case TESTMODE:
      // Test mode.  Encoder controls PWM rate for both plates
      if (runningTest) {
        for (int p = 0; p < PLATES; p++) {
          // Set the PWM outputs per plate based on the current encoder value
          ledcWrite(plates[p].pwmChannel, encPage);
          plates[p].ssrOut = encPage;
        }

        if (btnPressed) {
          // They pressed the button, so turn everything off and ramp up the middle fan
          for (int p = 0; p < 2; p++) {
            // Set the PWM outputs per plate based on the current encoder value
            ledcWrite(plates[p].pwmChannel, PWMOFF);
            plates[p].ssrOut = 0;
          }
          ledcWrite(coolingPwmChannel, COOLHIGH); 
          encPage = 0;
          runningTest = false;
          u8g2.setDrawColor(0);
          u8g2.drawBox(0, 16, 128, 8); 

          // Wait for the button to be released before proceeeding
          while (digitalRead(pPushButton) == BTNPRESSED) {
            delay(100);
          }
        }
      } else {
        if (btnPressed) {
          // They want to start the test, so setup the encoder variables, set the SSRs to the current encPage and set the middle fan to low
          u8g2.setDrawColor(0);
          u8g2.drawBox(0, 16, 128, 8); 
          runningTest = true;
          for (int p = 0; p < PLATES; p++) {
            // Set the PWM outputs per plate based on the current encoder value
            ledcWrite(plates[p].pwmChannel, encPage);
            plates[p].ssrOut = encPage;
          }
          ledcWrite(coolingPwmChannel, COOLLOW); 
          
          // Wait for the button to be released before proceeeding
          while (digitalRead(pPushButton) == BTNPRESSED) {
            delay(100);
          }
        }
      }

      // Show the testing state, the PWM value and the two plate temps
      u8g2.setDrawColor(1);
      u8g2.setFont(u8g2_font_6x10_tf);
      sprintf(outstr,"%s", runningTest ? "On" : "Off");
      u8g2.setDrawColor(0);
      u8g2.drawBox(64, 0, 20, 16);
      u8g2.setDrawColor(1);
      u8g2.drawStr(64, 16, outstr);
      sprintf(outstr,"%d / %d", (int)encPage, (int) PWMMAX);
      u8g2.setDrawColor(0);
      u8g2.drawBox(40, 24, 80, 8);
      u8g2.setDrawColor(1);
      u8g2.drawStr(40, 32, outstr);
      sprintf(outstr,"%dC   %d", (int)plates[1].curTemp, (int)plates[1].thermValue);
      u8g2.setDrawColor(0);
      u8g2.drawBox(40, 40, 80, 8);
      u8g2.setDrawColor(1);
      u8g2.drawStr(40, 48, outstr);
      sprintf(outstr,"%dC   %d", (int)plates[0].curTemp, (int)plates[0].thermValue);
      u8g2.setDrawColor(0);
      u8g2.drawBox(40, 56, 80, 8);
      u8g2.setDrawColor(1);
      u8g2.drawStr(40, 64, outstr);
      u8g2.sendBuffer();    
      break;        
  }

  //delay(100);
}
